/*
 * @(#)CreationTool.java
 *
 */
package de.renew.diagram;

import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.Drawing;
import CH.ifa.draw.framework.DrawingEditor;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.HJDError;
import CH.ifa.draw.standard.UndoableTool;

import de.renew.diagram.drawing.DiagramDrawing;

import java.awt.Cursor;
import java.awt.Point;
import java.awt.event.MouseEvent;


/**
 * This code is copied from CH.ifa.draw.standard and extended
 * to access functionality that was formerly private.
 * A tool to create new figures. The figure to be
 * created is specified by a prototype.
 *
 * <hr>
 * <b>Design Patterns</b><P>
 * <img src="images/red-ball-small.gif" width=6 height=6 alt=" o ">
 * <b><a href=../pattlets/sld029.htm>Prototype</a></b><br>
 * CreationTool creates new figures by cloning a prototype.
 * <hr>
 */
public class CreationTool extends UndoableTool {
    /**
     * Static logger for all instances of this class. 
     */
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(CreationTool.class);

    /**
     * The figure that is returned by the drawing
     * i.e. the figure that is drawn, if any
     * otherwise null.
     */
    private Figure fDrawnFigure = null;

    /**
    * the anchor point of the interaction
    */
    private Point fAnchorPoint;

    /**
     * the currently created figure
     */
    private Figure fCreatedFigure;

    /**
     * the prototypical figure that is used to create new figures.
     */
    private Figure fPrototype;

    /**
     * Initializes a CreationTool with the given prototype.
     * @param editor the DrawingEditor coordinator managing this tool
     * @param prototype the figure for which to work on
     */
    public CreationTool(DrawingEditor editor, Figure prototype) {
        super(editor);
        fPrototype = prototype;
    }

    /**
     * Constructs a CreationTool without a prototype.
     * This is for subclasses overriding createFigure.
     * @param editor the DrawingEditor coordinator managing this tool
     */
    protected CreationTool(DrawingEditor editor) {
        super(editor);
        fPrototype = null;
    }

    /**
     * Retrieve the figure which this tool is working on
     * @return the Figure
     */
    protected Figure getPrototype() {
        return fPrototype;
    }

    /**
     * Sets the cross-hair cursor.
     */
    @Override
    public void activate() {
        super.activate();
        view().setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
        fAnchorPoint = null;
        fCreatedFigure = null;
    }

    /**
     * Creates a new figure by cloning the prototype.
     * @param e the Mouse event which triggers this action, won't be used in the method
     * @param x the x coordinate
     * @param y the y coordinate
     */
    @Override
    public void mouseDown(MouseEvent e, int x, int y) {
        fAnchorPoint = new Point(x, y);
        fCreatedFigure = createFigure();
        fCreatedFigure.displayBox(fAnchorPoint, fAnchorPoint);
        fDrawnFigure = view().add(fCreatedFigure);
        changesMade();
    }

    /**
     * Creates a new figure by cloning the prototype.
     * @return a clone of the figure this CreationTool instance is working on
     * @throws HJDError if the figure defined for this tool is a Null at present
     */
    protected Figure createFigure() {
        if (fPrototype == null) {
            throw new HJDError("No prototype defined");
        }
        return (Figure) fPrototype.clone();
    }

    /**
     * Adjusts the extent of the created figure
     * @param e the triggering mouse event
     * @param x the x coordinate
     * @param y the y coordinate
     */
    @Override
    public void mouseDrag(MouseEvent e, int x, int y) {
        fCreatedFigure.displayBox(fAnchorPoint, new Point(x, y));
    }

    /**
     * Checks if the created figure is empty. If it is, the figure
     * is removed from the drawing.
     * See {@link Figure#isEmpty()}
     * @param e the triggering MouseEvent, won't be regarded
     * @param x the x coordinate, won't be regarded
     * @param y the y coordinate, won't be regarded
     */
    @Override
    public void mouseUp(MouseEvent e, int x, int y) {
        if (fCreatedFigure.isEmpty()) {
            drawing().remove(fCreatedFigure);
            noChangesMade();
        }
        fCreatedFigure = null;
        editor().toolDone();
    }

    /**
     * Gets the currently created figure
     * @return the created figure for this Tool
     */
    protected Figure createdFigure() {
        return fCreatedFigure;
    }

    /**
     * Gets the figure that is currently being drawn by the tool
     * @return the currently being drawn figure
     */
    public Figure getDrawnFigure() {
        return fDrawnFigure;
    }

    //NOTICEsignature

    /**
     * Determine and return the connector for the parent figure being worked on, using the x and y coordinates submitted.
     * If the parent is a {@link VSplitFigure}, the x value in relation to the parent's center x coordinate will determine where the
     * Connector is located on that parent. Otherwise, the connector will be located under the center of the parent at the edge of the parent's box.
     * @param parent the Parent in which to locate the Start connector
     * @param x the x coordinate to which the Connector points
     * @param y the y coordinate to which the Connector points
     * @return the determined and configured connector for the parent figure.
     */
    protected Connector determineStartConnector(Figure parent, int x, int y) {
        Connector connector = null;
        if (parent instanceof VSplitFigure) {
            if (x < parent.center().x) {
                connector = parent.connectorAt(parent.displayBox().x,
                                parent.center().y);
            } else {
                connector = parent.connectorAt(
                                parent.displayBox().x
                                                + parent.displayBox().width,
                                parent.center().y);
            }
        } else {
            connector = parent.connectorAt(parent.center().x,
                            parent.displayBox().y + parent.displayBox().height);
        }

        return connector;
    }

    /**
     * snap the position for a DiagramFigure so that the LifeLine is vertical
     * @param figure the Figure in whose context we are working
     * @param parent the parent for the submitted figure.
     * @param connector the connector we are working on
     */
    protected void snapToFit(Figure figure, Figure parent,
                             Connector connector) {
        if (!(parent instanceof RoleDescriptorFigure)) {
            int offset = 0;
            if (parent instanceof TaskFigure) {
                offset = parent.displayBox().width / 2;
            } else if (parent instanceof VSplitFigure) {
                if (connector instanceof RightConnector) {
                    offset = parent.displayBox().width;
                }
            } else if (parent instanceof VJoinFigure) {
                offset = parent.displayBox().width / 2;
            }

            figure.moveBy(parent.displayBox().x + offset - figure.displayBox().x
                            - figure.displayBox().width / 2, 0);
        }
    }

    /**
     * Attempt to locate the parent for a submitted figure. For the found parent and its {@link RoleDescriptorFigure}, membership
     * in that RoleDescriptionFigures hierarchy stack will be configured, and the head property of the figure will be configured.
     * If the parent is a {@link RoleDescriptorFigure} directly, a peer relationship with name "start" will be added to the submitted figure.
     * @param figure the figure whose parent to locate.
     * @return the located Parent figure of the submitted Figure.
     */
    protected Figure findParent(Figure figure) {
        logger.debug("1Figure  " + figure);
        TailFigure fig = (TailFigure) figure;
        Locator loc = getLocator();
        if (loc != null) {
            Figure dParent = loc.getElementAtPosition(fig.center().x,
                            fig.displayBox().y);
            logger.debug("2Figure Parent is " + dParent);
            if (dParent != null) {
                //loc.add(fig);
                RoleDescriptorFigure head = ((TailFigure) dParent).getDHead();
                if (head != null) {
                    head.addToTail(figure);
                    fig.setDHead(head);
                }
                TailFigure parent = (TailFigure) dParent;

                fig.addDParent(parent);
                parent.addDChild(fig);

                if (dParent instanceof RoleDescriptorFigure) {
                    fig.addPeerName("start");
                }

                //logger.debug("DiagramDrawing second "+ generatePeerInstantaneously);
                if (generatePeerInstantaneously()) {
                    (fig).generatePeers();
                }
                logger.debug("Figure added as Parent " + dParent);
                return dParent;
            }
        }
        return null;
    }

    /**
     * For the drawing the tool works on, if it is a DiagramDrawing, return the {@link Locator}
     * @return the locator for the DiagramDrawing this tool works on
     */
    protected Locator getLocator() {
        Drawing drawing = view().drawing();
        if (!(drawing instanceof DiagramDrawing)) {
            return null;
        }

        DiagramDrawing ddrawing = (DiagramDrawing) drawing;
        return ddrawing.getLocator();

    }

    /**
     * Check if the Peer relationship for this drawing can be generated instantaneously or not.
     * @return boolean True if instantaneous generation is possible. False otherwise.
     */
    protected boolean generatePeerInstantaneously() {
        Drawing drawing = view().drawing();
        if (!(drawing instanceof DiagramDrawing)) {
            return false;
        }

        return ((DiagramDrawing) drawing).generatePeerInstantaneously;

    }
}