/*
 * @(#)MessageConnection.java 5.1
 *
 */
package de.renew.diagram;

import CH.ifa.draw.figures.ArrowTip;
import CH.ifa.draw.figures.LineConnection;
import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureEnumeration;

import CH.ifa.draw.framework.Handle;
import CH.ifa.draw.standard.ChangeConnectionEndHandle;
import CH.ifa.draw.standard.ChangeConnectionStartHandle;
import de.renew.gui.AssocArrowTip;
import de.renew.gui.InscribableFigure;
import de.renew.shadow.ShadowNet;
import de.renew.shadow.ShadowNetElement;

import java.awt.Point;
import java.util.Vector;


// Added AssocArrowTip to the package 
//import de.renew.gui.fs.AssocArrowTip;

/**
 * MessageConnection implements the AIP semantics of message connections between different object activities
 * on other LifeLines. They are first-class distinct from {@link LifeLineConnection} as a consequence. MessageConnections
 * may be one of multiple types, configured and expressed via a ShadowNet semantic inscription according to
 * {@link de.renew.shadow.ShadowArc}.
 *
 */
public class MessageConnection extends LineConnection
                implements IDiagramElement, InscribableFigure {
    /**
     * Creates a new default MessageConnection with an end tip of type {@link AssocArrowTip}.
     */
    public MessageConnection() {
        super();
        ArrowTip start = null;
        ArrowTip end = new AssocArrowTip();

        setStartDecoration(start);
        setEndDecoration(end);
        setAttribute("ArrowTip", AssocArrowTip.class.getName());

    }

    /** Creates a new MessageConnection of a specific type.
     * @param arcType The type is defined according to
     *                the constants in
     *                {@link de.renew.shadow.ShadowArc}. Parameter is ignored in current implementation.
     */
    public MessageConnection(int arcType) {
        ArrowTip start = null;
        ArrowTip end = null;


        //setLineStyle(LINE_STYLE_DASHED);
        //start = new ArrowTip();
        end = new AssocArrowTip();

        setStartDecoration(start);
        setEndDecoration(end);
        setAttribute("ArrowTip", AssocArrowTip.class.getName());


    }

    //    public int getArcType() {
    //        int arrowMode = ((Integer) getAttribute("ArrowMode")).intValue();
    //        if (arrowMode == ARROW_TIP_NONE) {
    //            return ShadowArc.test;
    //        }
    //        if (arrowMode == ARROW_TIP_BOTH) {
    //            return ShadowArc.both;
    //        }
    //        return ShadowArc.ordinary;
    //    }
    @Override
    public boolean canConnect(Figure start, Figure end) {
        return (start instanceof TaskFigure && end instanceof TaskFigure
                        || start instanceof TaskFigure
                                        && end instanceof HSplitFigure
                        || start instanceof HSplitFigure
                                        && end instanceof TaskFigure
                        || start instanceof HSplitFigure
                                        && end instanceof HSplitFigure);
    }

    //    public boolean getTraceMode() {
    //        Object value = getAttribute("TraceMode");
    //        if (value instanceof Boolean) {
    //            return ((Boolean) value).booleanValue();
    //        }
    //        return true;
    //    }

    /**
     * Checks whether this MessageConnection is considered reversed or not.
     * @return Boolean True if connection is configured in reverse, Boolean False if not.
     */
    public boolean isReverse() {
        return ((Integer) getAttribute("ArrowMode"))
                        .intValue() == ARROW_TIP_START;
    }

    /**
     * For the figures this MessageConnection is linked with, add the appropriate Peer names to these figures
     * for eventual {@link de.renew.diagram.peer.DrawingPeer} generation under inclusion of the semantics of
     * this connection.
     */
    public void updatePeerNames() {
        Figure start = startFigure();
        if (start instanceof TaskFigure) {
            ((TaskFigure) start).addPeerName("out");
        }
        Figure end = endFigure();
        if (end instanceof TaskFigure) {
            ((TaskFigure) end).addPeerName("in");
        }
    }

    /**
     * For the figures this MessageConnection is linked to, create a new MessageConnection with reference
     * to this MessageConnection object.
     */
    public void notifyConnectorOwners() {
        Figure start = (startFigure());
        if (start instanceof TaskFigure) {
            ((TaskFigure) start).addMessage(this);
        }
        Figure end = endFigure();
        if (end instanceof TaskFigure) {
            ((TaskFigure) end).addMessage(this);
        }
    }

    /**
     * For the submitted figure, attempt to retrieve the connector for that figure.
     * @param fig The figure for which to retrieve the other connection end. Should be a figure this MessageConnection is linked with.
     * @return Either the connector at the start of the figure, or if the submitted figure owns the starting connector, the connector at the other end of the
     * MessageConnection.
     */
    public Connector getConnectedConnector(Figure fig) {
        Connector connector = start();
        if (!connector.owner().equals(fig)) {
            connector = end();
        }

        return connector;
    }

    /**
     * For the submitted figure, get either the opposite figure's end point, or the point for the link of this starting figure.
     * @param fig the figure for which to retrieve the opposite link coordinate
     * @return x,y canvas coordinates where the connector for the opposite figure's connector is located.
     */
    public Point getConnectedPoint(Figure fig) {
        Figure figure = startFigure();
        Point point = startPoint();
        if (!figure.equals(fig)) {
            point = endPoint();
        }
        return point;
    }

    /**
     * Retrieve the message text annotation on this MessageConnection.
     * @return String content of the {@link DiagramTextFigure} on this MessageConnection. Null if no annotation is present.
     */
    public String getMessageText() {
        String text = null;
        FigureEnumeration enumeration = children();
        while (enumeration.hasMoreElements()) {
            Figure textFigure = enumeration.nextElement();
            if (textFigure instanceof DiagramTextFigure) {
                DiagramTextFigure dtf = (DiagramTextFigure) textFigure;
                text = dtf.getRealText();
            }
        }
        return text;

    }

    @Override
    public ShadowNetElement buildShadow(ShadowNet net) {
        // TODO Auto-generated method stub
        return null;
    }
    
    @Override
    public ShadowNetElement getShadow() {
        // TODO Auto-generated method stub
        return null;
    }

    /**
     * This class provides a ChangeConnectionHandle for the start of a MessageConnection.
     * It ensures the right findConnector method is called.
     *
     * @author Lukas Seifert
     */
    private static class MessageChangeConnectionStartHandle extends ChangeConnectionStartHandle {
        /**
         * Constructs a handle with the given owner, locator, and connection prototype
         *
         * @param owner the owner of the connection
         */
        private MessageChangeConnectionStartHandle(Figure owner) {
            super(owner);
        }

        @Override
        protected Connector findConnector(int x, int y, Figure f) {
            if (f instanceof TaskFigure) {
                return ((TaskFigure) f).messageConnectorAt();
            }
            else {
                return f.connectorAt(x,y);
            }
        }
    }

    /**
     * This class provides a ChangeConnectionHandle for the end of a MessageConnection.
     * It ensures the right findConnector method is called.
     *
     * @author Lukas Seifert
     */
    private static class MessageChangeConnectionEndHandle extends ChangeConnectionEndHandle {
        /**
         * Constructs a handle with the given owner, locator, and connection prototype
         *
         * @param owner the owner of the connection
         */
        private MessageChangeConnectionEndHandle(Figure owner) {
            super(owner);
        }

        @Override
        protected Connector findConnector(int x, int y, Figure f) {
            if (f instanceof TaskFigure) {
                return ((TaskFigure) f).messageConnectorAt();
            }
            else {
                return f.connectorAt(x,y);
            }
        }
    }

    @Override
    public Vector<Handle> handles() {
        int n = fPoints.size();
        Vector<Handle> handles = new Vector<>(2 * n);
        handles.addElement(new MessageChangeConnectionStartHandle(this));
        addIntermediateHandles(handles);
        handles.addElement(new MessageChangeConnectionEndHandle(this));
        return handles;
    }

}