/*
 * @(#)RoleDescriptionFigure.java
 *
 */
package de.renew.diagram;

import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.Drawing;
import CH.ifa.draw.framework.DrawingView;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureEnumeration;
import CH.ifa.draw.framework.Handle;
import CH.ifa.draw.standard.FigureEnumerator;
import CH.ifa.draw.standard.MergedFigureEnumerator;

import de.renew.diagram.peer.DrawingPeer;
import de.renew.gui.CPNDrawing;
import de.renew.gui.GuiPlugin;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.Iterator;
import java.util.Vector;

/**
 * A RoleDescriptorFigure is a specific individual Role in an AIP drawing.
 * It is a special {@link TailFigure} and serves as the "head" of the hierarchies. It is the beginning of a lifeline.
 * It implements a special templating role and has a directly assigned name, which is represented in their
 * graphical drawing directly when produced through {@link DrawingPeer}.
 *
 * @author Lawrence Cabac
 */
public class RoleDescriptorFigure extends TailFigure {

    /**
     * This figure will be highlighted in an instance
     * drawing in the same manner as the transition
     * will be highlighted. May be <code>null</code>.
     * @serial
     **/
    private Figure hilightFig = null;
    /**
     * the drawing this RoleDescriptionFigure works on
     */
    private Drawing connectedDrawing;
    /**
     * The drawingView this Figure works on
     */
    private DrawingView connectedDrawingView;
    /**
     * Vector of all the Diagram figures which are part of the lifeline hierarchy stack of this figure.
     */
    private Vector<Figure> tail;
    /**
     * The name of the RoleDescriptorFigure
     */
    private String name;
    /**
     * The number of RoleDescriptorFigures which have been instantiated in the current running session 
     */
    private static int number = 0;

    /**
     * Produces a default name, which is "Role 1".
     *
     * @return the default name, which is "Role 1"
     */
    static String getDefaultName() {
        return "Role" + (++number);
    }

    /**
     * Create a new RoleDescriptorFigure. This is a basic TailFigure with the extension
     * that it tracks an (at instantiation empty) vector of tail objects, and assumes
     * its default name of "Role 1".
     */
    public RoleDescriptorFigure() {
        super();
        tail = new Vector<>();
        setName(getDefaultName());
    }

    @Override
    public Rectangle displayBox() {
        return new Rectangle(fDisplayBox.x, fDisplayBox.y, fDisplayBox.width,
                        fDisplayBox.height);
    }

    @Override
    protected void basicMoveBy(int x, int y) {
        fDisplayBox.translate(x, y);
    }

    /**
     * Changes the display box of a figure. Clients usually
     * call this method. It changes the display box
     * and announces the corresponding change.
     * @param origin the new origin
     * @param corner the new corner
     * {@link #displayBox}
     */
    @Override
    public void displayBox(Point origin, Point corner) {
        willChange();


        //basicDisplayBox(origin,new Point ( (int)(defaultDimension().getWidth() ),corner.y - origin.y));
        basicDisplayBox(origin,
                        new Point(corner.x - origin.x, corner.y - origin.y));
        changed();
    }

    @Override
    public void basicDisplayBox(Point origin, Point corner) {
        fDisplayBox = new Rectangle(origin.x, origin.y, corner.x, corner.y);
    }

    @Override
    public Connector connectorAt(int x, int y) {
        return new VerticalConnector(this);
    }

    @Override
    public Vector<Handle> handles() {
        Vector<Handle> handles = super.handles();
        handles.addElement(new LifeLineConnectionHandle(this));
        return handles;
    }

    /**
     * Returns whether the TraceMode is true or false.
     * Will return true, if the TraceMode is not an attribute.
     * @return boolean, depending on the TraceMode attribute
     */
    public boolean getTraceMode() {
        Object value = getAttribute("TraceMode");
        if (value instanceof Boolean) {
            return ((Boolean) value).booleanValue();
        }
        return true;
    }

    @Override
    public void release() {
        super.release();

    }

    /**
     * Creates a new Default Dimension with 72 as width and 36 as height.
     * @return a new Dimension
     */
    public static Dimension defaultDimension() {
        return new Dimension(72, 36);
    }

    /**
     * Sets a figure as the highlighted figure.
     * @param fig the highlighted figure
     */
    public void setHighlightFigure(Figure fig) {
        hilightFig = fig;
    }

    /**
     * Return the highlighted figure.
     * @return the highlighted figure
     */
    public Figure getHighlightFigure() {
        return hilightFig;
    }

    /**
     * Returns all figures with dependencies of the superclass
     * plus an optional hilight figure.
     **/
    @Override
    public FigureEnumeration getFiguresWithDependencies() {
        FigureEnumeration superDep = super.getFiguresWithDependencies();
        Vector<Figure> myDep = new Vector<>(1);
        myDep.addElement(getHighlightFigure());
        return new MergedFigureEnumerator(superDep,
                        new FigureEnumerator(myDep));
    }

    /**
     *  Starts the generation of Peers, if a drawing exists.
     *  Uses createNewConnectedDrawing to get the drawing from the GUI
     *  {@link #createNewConnectedDrawing()}
     */
    public void createNewDrawingAndGeneratePeers() {
        updateName();
        if (!getName().substring(0, 1).equals("!")) {
            createNewConnectedDrawing();
            if (getConnectedDrawing() != null) {
                startGeneratingPeers();
            } else {
                logger.error("Couldn't generate peers because no drawing exists!");
            }
        }
    }

    /**
     *  Gets the current Drawing and DrawingView from the GuiPlugin.
     */
    protected void createNewConnectedDrawing() {
        GuiPlugin starter = GuiPlugin.getCurrent();
        if (starter != null) {
            Drawing drawing = new CPNDrawing();
            updateName();
            drawing.setName(getName());
            drawing = starter.getGui().openDrawing(drawing);
            if (drawing != null) {
                setConnectedDrawing(drawing);
                setConnectedDrawingView(starter.getView(drawing));
            }
        }
    }

    private void updateName() {
        Iterator<Figure> it = children.iterator();
        while (it.hasNext()) {
            Figure fig = it.next();
            if (fig instanceof DiagramTextFigure) {
                setName(((DiagramTextFigure) fig).getText());
                return;
            }
        }
    }

    /**
     * Descending through the hierarchy of all children of this Head, begin building
     * Peers along the hierarchy of the elements - parents will have peers generated
     * before their children.
     */
    protected void startGeneratingPeers() {
        // fifo queue for putting all children in to generate peers
        // be careful not to put duplicates into the queue
        Vector<Figure> fifo = new Vector<>(getDCildren());
        // alreadyDone collects all figures that are completed
        Vector<Figure> alreadyDone = new Vector<>();
        alreadyDone.add(this);
        while (!fifo.isEmpty()) {
            Object o = fifo.remove(0);
            TailFigure child = (TailFigure) o;
            // checks if the parents were already done;
            // if this is not the case the figure is added to the end of fifo again.
            if(!alreadyDone.containsAll(child.getDParents())){
                fifo.add(child);
            }
            else {
                Vector<Figure> children = child.getDCildren();
                //ensure that all duplicates are removed
                children.removeAll(fifo);
                fifo.addAll(children);
                child.generatePeers();
                alreadyDone.add(child);
            }
        }

    }

    /**
     * Return the drawing object this RoleDescriptionFigure is in a relation with.
     * @return the Drawing in the relationship
     */
    public Drawing getConnectedDrawing() {
        return connectedDrawing;
    }

    /**
     * Set the drawing object this RoleDescriptionFigure should be in a relation with.
     * @param drawing the drawing this figure relates to.
     */
    public void setConnectedDrawing(Drawing drawing) {
        connectedDrawing = drawing;
    }

    /**
     * Add a Figure to the tail of this figure.
     * @param fig the Figure to add to the tail
     */
    public void addToTail(Figure fig) {
        tail.add(fig);
    }

    /**
     * Delete a figure from the tail of this figure.
     * @param fig the Figure to delete from the tail
     */
    public void removeFromTail(Figure fig) {
        tail.remove(fig);
    }

    /**
     * Retrieve an iterator to work over the Tail vector of this RoleDescriptionFigure
     * @return the iterator working on the tail storage
     */
    public Iterator<Figure> tail() {
        return tail.iterator();
    }

    // ------------------ TailFigure --------------------------------


    /**
     * RoleDescriptionFigures are their own head, and thus return their own reference on
     * calls for their hierarchy head.
     * @return identity of this object
     */
    @Override
    public RoleDescriptorFigure getDHead() {
        return this;
    }

    /**
     * Setting the roleDescriptionFigure is not supported on RoleDescriptionFigures themselves.
     * Any such inputs are ignored
     * @param rd the RoleDescriptionFigure to set; will be ignored.
     */
    @Override
    public void setDHead(RoleDescriptorFigure rd) {
    }


    /**
     * RoleDescriptionFigures are the heads of their parents and do not provide
     * capability to add any parents above them in the hierarchy.
     * @param parent the TailFigure which will become a parent of this TailFigure; will be ignored.
     */
    @Override
    public void addDParent(TailFigure parent) {
        // impl (or maybe not, cause there is no parent to this element)
    }


    // -------------------

    /**
     * TailDescriptionFigures have no peer listing and will always return null on query for such.
     * @return Null
     */
    @Override
    public Vector<DrawingPeer> getPeers() {
        return null;
    }

    /**
     * Return the drawing view this RoleDescriptionFigure is working with.
     * @return the DrawingView
     */
    public DrawingView getConnectedDrawingView() {
        return connectedDrawingView;
    }

    /**
     * Configure a DrawingView to work with for this RoleDescriptionFigure
     * @param view the View to set
     */
    public void setConnectedDrawingView(DrawingView view) {
        connectedDrawingView = view;
    }

    /**
     * Return the string name for this RoleDescriptionFigure, naming the Role in the AIP drawing.
     * @return the String name of this RoleDescriptionFigure.
     */
    public String getName() {
        return name;
    }

    /**
     * Set a new string name for the Role in the AIP drawing for this figure
     * @param string the String name to use.
     */
    public void setName(String string) {
        name = string;
    }

    /**
     * Since RoleDescriptorFigures don't manage peers, they do not update any peer names.
     */
    @Override
    public void updatePeerNames() {
        // nothing to do
    }

    /**
     * Flush all stored references to tail members from the tail and reset to nothing stored.
     */
    public void flushTail() {
        tail = new Vector<>();
    }

    /**
     * For all connections this RoleDescriptorFigure holds with children objects,
     * find those children, add them to this Figure's children, configure
     * this figure as those children's parents, add a parent connector with
     * this figure, and then have the children notify their parents in turn, with
     * this figure as the Head to use.
     */
    @Override
    public void findAndNotifyDChildren() {
        Iterator<LifeLineConnection> it = connections.iterator();
        while (it.hasNext()) {
            LifeLineConnection connection = it.next();
            TailFigure dChild = connection.otherEnd(this);


            logger.debug("=========>RoleDescriptor found!");

            addDChild(dChild);
            dChild.addDParent(this);
            dChild.addDParentConnector(getConnectedConnector(connection));
            dChild.notifyDesendent(this, getDHead());

        }
    }
    
    @Override
    protected void notifyDesendent(TailFigure parent,
                                   RoleDescriptorFigure head) {
        setDHead(head);

        findAndNotifyDChildren();
        getDHead().addToTail(this);

    }
}