/*
 * Created on Apr 29, 2003
 *
 */
package de.renew.diagram;

import CH.ifa.draw.DrawPlugin;
import CH.ifa.draw.application.DrawApplication;
import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.Drawing;
import CH.ifa.draw.framework.DrawingView;
import CH.ifa.draw.framework.Figure;

import de.renew.diagram.drawing.DiagramDrawing;
import de.renew.diagram.peer.DrawingPeer;
import de.renew.diagram.peer.IDrawingPeer;

import java.awt.Point;
import java.util.Iterator;
import java.util.Vector;


/**
 *<p>
 * A TailFigure Element of a DiagramDrawing is an element that can be added to a
 * tail; i.e. belongs to and is known by a RoleDescriptorFigure, which is the TailFigure's head. The TailFigure knows its
 * head (the managing RoleDescriptorFigure) and its parents (the elements it is connected to that is closer to the
 * head) as well as its children (the elements it is connected to that are further from the head).
 *</p>
 * <p>
 * TailFigures can compute their position on the canvas relative to other TailFigures they are in relation with, and
 * are also responsible for managing their connector objects and connector relations with other objects.
 * </p>
 * <p>
 * This class provides default implementation of some methods and stub implementations for override by inheritors.
 * </p>
 *
 * @author Lawrence Cabac
 */
public class TailFigure extends DiagramFigure {
    /**
     * Logger instance for all the instances of the class
     */
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(TailFigure.class);
    /**
     * Vector of all LifeLineConnections with other Figures through lifelines in the AIP diagram. 
     */
    protected Vector<LifeLineConnection> connections;
    /**
     * Default start location for a TailFigure at coordinates (50,500).
     */
    static final Point START_LOCATION = new Point(50, 500);
    /**
     * The RoleDescriptorFigure managing the stack of LifeLine-connected figures this TailFigure is a part of.
     */
    private RoleDescriptorFigure head;
    /**
     * The Vector of Peer objects which represents this TailFigure as a Petri Net subnet in the larger drawing that will be produced
     */
    protected Vector<DrawingPeer> drawingPeers; ///%%%%%
    /**
     * The Vector of Peer string names. Peer string names are identical with the template name that will be used to generate the Peer object.
     */
    private Vector<String> peerNames;
    /**
     * Vector of parent figures which sit "upstream" in the hierarchy of TailFigures.
     */
    private Vector<Figure> dParents;
    /**
     * Vector of Connector objects connecting this figure with its parent(s)
     */
    private Vector<Connector> dParentConnectors;
    /**
     * Vector of child figures this figure is in relation with
     */
    private Vector<Figure> dChildren;

    /**
     * Create a new TailFigure instance.
     */
    public TailFigure() {
        super();
        dParents = new Vector<>();
        dChildren = new Vector<>();
        peerNames = new Vector<>();
        drawingPeers = new Vector<>();
        dParentConnectors = new Vector<>();
        connections = new Vector<>();
    }

    /**
     * Retrieve the Head of the TailFigure, which is its managing RoleDescriptionFigure. If the Head is not
     * explicitly configured, the function attempts to retrieve the closest figure north of this Figure's center, and query
     * that figure for its head to set as this Figure's head.
     * @return the Head of this object's RoleDescriptionFigure. If the recursive search for a valid RoleDescriptionFigure fails, may return a null.
     */
    public RoleDescriptorFigure getDHead() {
        if (head == null) {
            Vector<Figure> v = getDParents();
            if (v.size() > 0) {
            } else {
            }

            DrawApplication gui = DrawPlugin.getGui();
            if (gui != null) {
                Locator loc = getLocator(gui.view());
                Figure fig = loc.getElementAtPosition(center().x,
                                displayBox().y);
                if (fig instanceof TailFigure) {
                    TailFigure tail = (TailFigure) fig;
                    setDHead(tail.getDHead());
                }
            }
        }
        return head;
    }

    /**
     * Set the RoleDescriptionFigure for this TailFigure.
     * @param rd the RoleDescriptionFigure to set
     */
    public void setDHead(RoleDescriptorFigure rd) {
        head = rd;
    }

    /**
     * Add the submitted TailFigure as a parent of this Figure.
     * @param parent the TailFigure which will become a parent of this TailFigure
     */
    public void addDParent(TailFigure parent) {
        dParents.add(parent);
    }

    /**
     * Check whether the TailFigure has any parents at all. Note that the RoleDescriptionFigure is not considered a parent.
     * @return True if there are any parents set. False otherwise.
     */
    public boolean hasDParent() {
        return dParents != null && dParents.size() > 0;
    }

    /**
     * Retrieve the Vector of all Diagram Figures which have been configured as parents of this Figure.
     * @return downcast Vector of all TailFigures set as parents of this TailFigure
     */
    public Vector<Figure> getDParents() {
        return dParents;
    }

    /**
     * Add a new Connector entity for some parent Figure of this TailFigure. Children manage the connectors to their parents only. Connectors
     * for a child should not be submitted here; instead add a connector to that child.
     * @param connector the Connector to anchor graphics from the parent.
     */
    public void addDParentConnector(Connector connector) {
        dParentConnectors.add(connector);
    }

    /**
     * Retrieve the Vector of Connector objects for parent objects of this TailFigure.
     * @return the Vector listing of all Connector objects for parents
     */
    public Vector<Connector> getDParentConnectors() {
        return dParentConnectors;
    }

    /**
     * Add a new Child TailFigure to the relations of this Figure.
     * @param child the TailFigure which will become a child relative of this Figure
     */
    public void addDChild(TailFigure child) {
        dChildren.add(child);
    }

    /**
     * Retrieve the Vector of all TailFigures which have child relations with this Figure
     * @return the Vector listing of all TailFigure children of this Figure
     */
    public Vector<Figure> getDCildren() {
        return dChildren;
    }

    /**
     * Add a new DrawingPeer object to his TailFigure's storage
     * @param peer the DrawingPeer Object which this object is in relation with
     */
    public void addPeer(DrawingPeer peer) {
        drawingPeers.add(peer);
    }

    /**
     * Retrieve the Vector of Peers objects in relation with this TailFigure
     * @return the Vector of DrawingPeers in relation with this object
     */
    public Vector<DrawingPeer> getPeers() {
        logger.debug("drawingPeers are " + drawingPeers);

        return drawingPeers;
    }

    /**
     * Update all Peers currently configured for this Figure, then execute Draw Calls for all the Peers and connect
     * the resulting Peer drawings together.
     */
    public void generatePeers() {
        updatePeers();


        // for all peers of one figure
        Iterator<DrawingPeer> it = getPeers().iterator();
        while (it.hasNext()) {
            IDrawingPeer peer = it.next();
            logger.debug("2222" + getDHead().getConnectedDrawing());
            peer.drawFigures(getDHead().getConnectedDrawingView());
        }
        it = getPeers().iterator();
        while (it.hasNext()) {
            IDrawingPeer peer = it.next();
            peer.connectFigures(getDHead().getConnectedDrawingView());
        }
    }

    // from the peer name list all peers are being produced.
    // some elements e.g. TaskFigures can hold more than one peer

    /**
     * Using String Name representations of peer names currently configured for this figure, generate DrawingPeers
     * of those names with the current Figure as the owner.
     */
    public void updatePeers() {
        drawingPeers = new Vector<>();
        Iterator<String> it = peerNames.iterator();
        while (it.hasNext()) {
            String name = it.next();
            if (name.equals("start")) {
                addPeer(new DrawingPeer(START_LOCATION, name + ".rnw", this,
                                this));
            } else {
                addPeer(name + ".rnw");
            }
        }
    }

    /**
     * Retrieve the next (x,y) location, defined as a coordinate on the last Peers first Figure. See {@link DrawingPeer} concept of the nextLocations.
     * @return (x,y) coordinates relative to the Last Peer's first Figure member.
     */
    public Point getNextLocation() {
        Point result = null;
        DiagramFigure parent = (DiagramFigure) getDParents().firstElement();
        if (parent != null) {
            result = lastPeer().getNextLocations().firstElement();
            // fix this to be able to access other Elements
        }
        return result;
    }

    //    public Point getNextLocation() {
    //        Point result = null;
    //        AttributeFigure fig = null;
    //        DiagramFigure parent = (DiagramFigure) getDParents().firstElement();
    //        if (parent != null && parent instanceof VSplitFigure){
    //            Connector connector = (Connector) dParentConnectors.firstElement();
    //            Vector endFigures =
    // ((DrawingPeer)((TailFigure)parent).getPeers().firstElement()).getEndFigures();
    //            
    //            if (connector instanceof LeftBottomConnector){
    //                fig =(AttributeFigure)endFigures.firstElement();
    //           }
    //            else {
    //                fig =(AttributeFigure)endFigures.lastElement();
    //            }
    // 
    //        }else{
    //            Vector endFigures =
    // ((DrawingPeer)((TailFigure)parent).getPeers().firstElement()).getEndFigures();
    //            fig =(AttributeFigure)endFigures.firstElement();
    //        }
    //        result.x = fig.center().x + 100;
    //        result.y = fig.center().y;
    //
    //        //result = (Point) lastPeer().getNextLocations().firstElement(); // fix
    // this to be able to access other Elements
    //        return result;
    //    }


    /**
     * Attempt to retrieve the last DrawingPeer in the list of this instance's peers. If no peer has been configured,
     * retrieve the last peer of the first parent.
     * @return The last Peer that is in the list of Peers. If no parent is defined, the last peer of the
     * first Parent is defined as the last Peer. Null if no Peer can be found.
     */
    public IDrawingPeer lastPeer() {
        Vector<DrawingPeer> v = getPeers();
        logger.debug("2nd drawingPeers are " + v + "  " + this);

        if (v != null) {
            if (v.isEmpty()) {
                //     logger.debug("peerlist empty!");
                return ((TailFigure) getDParents().firstElement()).lastPeer();
            }
            logger.info(getPeers().lastElement());
            logger.info(getPeers().firstElement());
            return getPeers().lastElement();
        }
        return null;
    }

    /**
     * For a given Peer, retrieve the preceding Peer of that peer in the vector stack.
     * @param peer the peer for which to get the preceding peer
     * @return the preceding peer of this object, if it exists. if the submitted peer is the first peer of ths object,
     * query the preceding figure for its last peer and return that result.
     */
    public IDrawingPeer getPrecedingPeer(IDrawingPeer peer) {
        int index = drawingPeers.indexOf(peer);
        if (index == 0) {
            // first peer just return the last of the preceding tailable Figure
            return ((TailFigure) getDParents().firstElement()).lastPeer();
        }
        return drawingPeers.elementAt(index - 1);
    }

    // FIXME fix this later
    //NOTICEsignature
    /**
     * Retrieve the second peer before this peer. CAUTION: this function is broken at the moment.
     * @param peer the peer for which to get the second-preceding peer
     * @return the second preceding peer of this object, if it exists.
     */
    public IDrawingPeer getSecondPrecedingPeer(IDrawingPeer peer) {
        TailFigure parent = ((TailFigure) getDParents().elementAt(1));
        IDrawingPeer dp = parent.lastPeer();
        return dp;

    }

    /**
     * Remove all Peer objects from this TailFigure.
     */
    public void removePeers() {
        drawingPeers = new Vector<>();

    }

    /**
     * Add a new Peer object by its String name
     * @param name the String name of the Peer
     */
    public void addPeerName(String name) {
        peerNames.add(name);
    }

    /**
     * Adds a new peer with the given name to the existing peers.
     * First the location of the peer is calculated which is the point, the center of the peer is placed at.
     *
     * @param fileName The name of the peer to be added
     */
    public void addPeer(String fileName) {
        Point loc = calculateLocation(); // has to be changed soon
        logger.debug("------------------------------------------------- ");
        logger.debug(" ");

        addPeer(new DrawingPeer(loc, fileName, this, this));
        logger.debug("(((((((((((((((((((((((((((((((((((((((((((((( ");
        logger.debug("TailableFigure " + this);
        logger.debug("Parent " + getDParents().firstElement());
        logger.debug("Head " + getDHead());
        //        logger.debug("Parent Connector " +
        // getDParentConnectors().firstElement());
        logger.debug(")))))))))))))))))))))))))))))))))))))))))))))) ");

    }

    /**
     * Remove all Peer Objects and all Peer names from this TailFigure
     */
    public void flushPeers() {
        peerNames = new Vector<>();
        drawingPeers = new Vector<>();
    }

    /**
     * Remove all parent and child TailFigures and their Connectors from this TailFigure
     */
    public void flushRelatives() {
        dParents = new Vector<>();
        dParentConnectors = new Vector<>();
        dChildren = new Vector<>();
    }

    /*
     * Calculates the location the center of the peer is placed at.
     * Elements in a sequence have the same height.
     * After a split the elements connecting to it are placed at the same height as the connections.     *
     * In a Join the height is calculated as the middle of the height of the elements connecting to it.
     */

    /**
     * Calculates the location of the center of the peer for its placement in a larger figure. This calculation
     * is relative to other figures this TailFigure is in relation with. Elements in a sequence have the same height.
     * After a Split (see {@link VSplitFigure}) the elements connecting to the split are placed at the same height as the connections.
     * In a Join {@link VJoinFigure} the height of the element is calculated as the middle of the height of the elements connecting to it.
     * @return an x,y coordinate in Canvas origin for this TailFigure to be placed at.
     */
    protected Point calculateLocation() {
        if (this instanceof RoleDescriptorFigure) {
            return START_LOCATION;
        }
        Point p = null;
        try {
            Vector<Connector> dParentConnectors = getDParentConnectors();
            Vector<Figure> dPar = getDParents();
            logger.debug("this " + this);
            //        logger.debug("parentsConnectors "+v);
            logger.debug("parents " + dPar);
            logger.debug("parentsConnectors " + dParentConnectors);

            if (this instanceof VJoinFigure) {
                IDrawingPeer pd1 = ((TailFigure) dPar.firstElement())
                                .lastPeer();
                IDrawingPeer pd2 = ((TailFigure) dPar.elementAt(1)).lastPeer();
                Point p1 = pd1.getNextLocations().firstElement();
                Point p2 = pd2.getNextLocations().firstElement();
                p = new Point(Math.max(p1.x, p2.x), (p1.y + p2.y) / 2);

            } else {
                Connector con = dParentConnectors.firstElement();
                int index = 0;
                if (con instanceof RightConnector) {
                    index++;
                }
                logger.debug("TailFigure::calculateLocation line 344: this = "
                                + this);
                IDrawingPeer lastPeer = lastPeer();

                if (lastPeer != null) {
                    Vector<Point> nextLocations = lastPeer.getNextLocations();
                    try {
                        p = nextLocations.elementAt(index); //TODO: fix
                                                            // this index


                    } catch (ArrayIndexOutOfBoundsException e) {
                        // logger.error(e.getMessage(), e);
                        p = nextLocations.elementAt(0);
                    }
                }

                //        if (this instanceof TaskFigure && dParents.size() > 0){
                //            // if TaskFigure then there is only one dParent
                //            DiagramFigure parent = (DiagramFigure)
                // dParents.firstElement();
                //            if (drawingPeers.size()==0 && parent instanceof
                // VSplitFigure){
                //                
                //            }
                //        }
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            logger.error(e.getMessage(), e);
        }
        return p;
    }

    /**
     * Reset the list of Peer names for peer generation, erasing all previous Peer names.
     * Existing Peer object relations are not reset.
     */
    public void newPeerNames() {
        peerNames = new Vector<>();
    }

    /**
     * STUB update the Peer Names based on present configuration. This method is overridden
     * by inheritors of this default implementation
     */
    // Implementing classes override this method
    public void updatePeerNames() {
        //
    }

    /**
     * Add a new LifeLineConnection to this figure, connecting it with another figure.
     * @param connection the Connection to add to this Figure
     */
    public void addConnection(LifeLineConnection connection) {
        connections.add(connection);
    }

    /**
     * For a given connection object, retrieve the Connector that positions it on this TailFigure
     * @param connection the connection Object for which to retrieve the connector
     * @return the Connector for the given LifeLifeConnection. The end of the Connector if its startFigure is this
     * TailFigure; the start of the connector otherwise.
     */
    protected Connector getConnectedConnector(LifeLineConnection connection) {
        Figure fig = connection.startFigure();
        Connector connector;
        if (!fig.equals(this)) {
            fig = connection.endFigure();
            connector = connection.end();
        } else {
            connector = connection.start();
        }
        return connector;
    }

    /**
     * Delete all LifeLineConnection relations for this object.
     */
    public void flushConnections() {
        connections = new Vector<>();
    }

    /**
     * A TailFigure should update the structure of the peer components. and
     * notify its desendents to do likewise. Empty implementation for override
     * by inheritors.
     *
     * @param parent the identity of the parent from which the notification originates(?)
     * @param head the Head responsible for this chain of related TailFigures.
     */
    protected void notifyDesendent(TailFigure parent,
                                   RoleDescriptorFigure head) {
    }

    /**
     * Find all children of this figure and notify them of some event. Empty implementation for override
     * by inheritors.
     */
    public void findAndNotifyDChildren() {
    }

    private Locator getLocator(DrawingView view) {
        Drawing drawing = view.drawing();
        if (!(drawing instanceof DiagramDrawing)) {
            return null;
        }

        DiagramDrawing ddrawing = (DiagramDrawing) drawing;
        return ddrawing.getLocator();

    }
}