/*
 * @(#)TransitionFigure.java 5.1
 *
 */
package de.renew.diagram;

import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureEnumeration;
import CH.ifa.draw.framework.Handle;
import CH.ifa.draw.standard.FigureEnumerator;
import CH.ifa.draw.standard.MergedFigureEnumerator;
import CH.ifa.draw.util.StorableInput;
import CH.ifa.draw.util.StorableOutput;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;


/**
 * Splits a {@link LifeLineConnection} into two LifeLineConnections, with specific specialized behavior
 * depending on the set {@link FigureDecoration}.
 *
 * @author Lawrence Cabac
 */
public class VSplitFigure extends TailFigure implements ISplitFigure {
    /**
     * logger entity for all instances of this class.
     */
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(VSplitFigure.class);

    /**
     * Create a new VSplitFigure DiagramFigure.
     */
    public VSplitFigure() {
        super();
    }

    /**
     * Changes the display box of a figure. Clients usually
     * call this method. It changes the display box
     * and announces the corresponding change.
     * @param origin the new origin
     * @param corner the new corner
     */
    @Override
    public void displayBox(Point origin, Point corner) {
        willChange();
        basicDisplayBox(new Point(origin.x, origin.y), new Point(
                        corner.x - origin.x,
                        (int) (defaultDimension().getHeight())));
        changed();
    }

    /**
     * Locate connectors on the SplitFigure depending on which coordinates are submitted for the request.
     * Requesting a connection in the middle 2/4ths of the figure horizontal extend returns a new {@link VSplitCenterConnector}.
     * Requesting a connection left of the center 2/4ths returns a new {@link LeftConnector}, on the right a new {@link RightConnector}
     * is returned.
     * @param x the X coordinate in the local frame with origin at the displayBox top left
     * @param y the Y coordinate in the local frame with origin at the displayBox top left
     * @return the appropriately selected new connector object.
     */
    @Override
    public Connector connectorAt(int x, int y) {
        Rectangle r = displayBox();
        int left = r.x;
        int right = r.x + r.width;
        int length = right - left;

        int forth = (length / 4);
        if (x < left + forth) {
            return new LeftConnector(this);
        } else if (x > right - forth) {
            return new RightConnector(this);
        }

        return new VSplitCenterConnector(this);
    }

    /**
     * The default dimension of a VSplitFigure is considered to be a width of 55, height of 8.
     * @return the Dimension object encapsulating the default values.
     */
    public static Dimension defaultDimension() {
        return new Dimension(55, 8);
    }

    @Override
    public Vector<Handle> handles() {
        Vector<Handle> handles = new Vector<>();
        VSplitFigureBoxHandleKit.addHandles(this, handles);
        return handles;

    }

    /**
     * The FigureDecoration object of this class controls the semantic specialization of the
     * Split into different types (AND, OR, XOR) as well as visual appearance.
     */
    private FigureDecoration _decoration;
    /**
     * the type of the FigureDecoration is relevant for peer specialization.
     */
    protected Class<?> _decorationClass = FigureDecoration.class;

    @Override
    public Rectangle displayBox() {
        return new Rectangle(fDisplayBox.x, fDisplayBox.y, fDisplayBox.width,
                        fDisplayBox.height);
    }

    @Override
    protected void basicMoveBy(int x, int y) {
        fDisplayBox.translate(x, y);
    }

    /**
     * Changes the display box of a figure. Clients usually
     * call this method. It changes the display box
     * and announces the corresponding change.
     * @param origin the new origin
     * @param corner the new corner
     */
    @Override
    public void basicDisplayBox(Point origin, Point corner) {
        fDisplayBox = new Rectangle(origin.x, origin.y, corner.x, corner.y);
    }

    @Override
    public void internalDraw(Graphics g) {
        super.internalDraw(g);
        decorate(g);
    }

    @Override
    public void setDecoration(FigureDecoration fd) {
        _decoration = fd;
    }

    @Override
    public FigureDecoration getDecoration() {
        return _decoration;
    }

    @Override
    public boolean hasDecoration() {
        return true;
    }

    private void decorate(Graphics g) {
        if (_decoration != null) {
            int x = fDisplayBox.x + fDisplayBox.width / 2;
            int y = fDisplayBox.y + fDisplayBox.height / 2;


            //Point p4 = (Point) fPoints.elementAt(fPoints.size() - 1);
            _decoration.draw(g, x, y, getFillColor(), getFrameColor());
        }
    }

    /**
     * Ge the TraceMode attribute setting for this figure.
     * @return True by default. The boolean setting of the TraceMode attribute otherwise.
     */
    public boolean getTraceMode() {
        Object value = getAttribute("TraceMode");
        if (value instanceof Boolean) {
            return ((Boolean) value).booleanValue();
        }
        return true;
    }

    @Override
    public void release() {
        super.release();

    }

    /**
     * Returns all figures with dependencies of the superclass
     * plus an optional hilight figure.
     * @return the FigureEnumeration container of all the figures
     **/
    @Override
    public FigureEnumeration getFiguresWithDependencies() {
        FigureEnumeration superDep = super.getFiguresWithDependencies();
        Vector<Figure> myDep = new Vector<>(1);

        //myDep.addElement(getHighlightFigure());
        return new MergedFigureEnumerator(superDep,
                        new FigureEnumerator(myDep));
    }

    @Override
    public void write(StorableOutput dw) {
        super.write(dw);


        dw.writeStorable(_decoration);
        dw.writeString(_decorationClass.getName());
    }

    @Override
    public void read(StorableInput dr) throws IOException {
        super.read(dr);
        _decoration = (FigureDecoration) dr.readStorable();
        dr.readString();
    }


    /**
     * Depending on the type of {@link FigureDecoration} set for this figure,
     * set the peer specialization appropriately.
     */
    @Override
    public void updatePeerNames() {
        if (getDecoration() instanceof XORDecoration) {
            addPeerName("cond");
        }
        if (getDecoration() instanceof ANDDecoration) {
            addPeerName("psplit");
        }
    }


    /**
     * Notify descendant children of this VSplitFigure, which are those figures sitting beyond the LeftConnector
     * and RightConnector objects of this figure, that they are now in a LifeLife hierarchy stack, and set configurations
     * for those children appropriately.
     */
    @Override
    public void findAndNotifyDChildren() {
        Iterator<LifeLineConnection> it = connections.iterator();
        while (it.hasNext()) {
            LifeLineConnection connection = it.next();
            TailFigure dChild = connection.otherEnd(this);


            logger.debug("=========>VSplitFigure found!");

            Connector connector = getConnectedConnector(connection);
            if (connector instanceof LeftConnector
                            || connector instanceof RightConnector) {
                addDChild(dChild);
                dChild.addDParent(this);
                dChild.addDParentConnector(connector);
                dChild.notifyDesendent(this, getDHead());

            }
        }
    }

    /**
     * Notify the descendants of this figure about their relation in the Lifeline stack. Set the parent
     * for this figure.
     * @param parent ignored in this case
     * @param head the Head responsible for this chain of related TailFigures.
     */
    @Override
    protected void notifyDesendent(TailFigure parent,
                                   RoleDescriptorFigure head) {
        setDHead(head);
        findAndNotifyDChildren();
        getDHead().addToTail(this);

    }
}