package de.renew.diagram;

import CH.ifa.draw.framework.ConnectionFigure;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.standard.AbstractConnector;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;


/**
 * A VerticalConnector is a special connector used for vertical connections on {@link DiagramFigure}.
 * It locates connection points by choping the connection between the centers of the
 * two figures at the display box.
 *
 * {@link CH.ifa.draw.framework.Connector}
 * @author Lawrence Cabac
 */
public class VerticalConnector extends AbstractConnector {

    /**
     * Creates a new default VerticalConnector.
     */
    public VerticalConnector() { // only used for Storable implementation
    }

    /**
     * Creates a new VerticalConnector with an owner.
     *
     * @param owner The owner of the connector
     */
    public VerticalConnector(Figure owner) {
        super(owner);
    }

    @Override
    public Point findStart(ConnectionFigure connection) {
        Rectangle r1;
        if (connection.pointCount() == 2) {
            r1 = connection.end().displayBox();
        } else {
            r1 = new Rectangle(connection.pointAt(1), new Dimension(0, 0));
        }

        //NOTICEsignature
        return chop(connection.start().owner(), r1, connection.startPoint().y);

    }

    @Override
    public Point findEnd(ConnectionFigure connection) {
        Rectangle r1;
        if (connection.pointCount() == 2) {
            r1 = connection.start().displayBox();
        } else {
            r1 = new Rectangle(connection.pointAt(connection.pointCount() - 2),
                            new Dimension(0, 0));
        }

        //NOTICEsignature
        return chop(connection.end().owner(), r1, connection.startPoint().y);
    }

    /**
     * Computes a coordinate that "chops" the target figure entity's dimensions in half
     * on the horizontal axis under consideration of the relative position of the source entity.
     *
     * The computed coordinate will be depending on the source position either at the left or right
     * and top or bottom.
     *
     * @param target the target figure whose coordinate is to be returned
     * @param source the source coordinate relative to which the connection will be computed
     * @param xxx will be ignored
     * @return The computed coordinate snapped to the target figure
     */
    //NOTICEsignature
    protected Point chop(Figure target, Rectangle source, int xxx) {
        Rectangle r = target.displayBox();
        int h = (int) (r.getHeight());
        int off = (int) (source.getWidth() / 2);
        int horizontal = source.x + off;
        int vertical = r.y;


        if (source.x + off < r.x) {
            horizontal = r.x;
        } else if (source.x + off > r.x + r.getWidth()) {
            horizontal = (int) (r.x + r.getWidth());
        }
        if (source.y > r.y) {
            vertical = r.y + h;
        }
        return new Point(horizontal, vertical);


    }

    /**
    * The display box of a <code>VerticalConnector</code> is a
    * zero-height rectangle covering the top of the owner figure.
    *
    * @return the top rectangle of the owner figure.
    **/
    @Override
    public Rectangle displayBox() {
        Rectangle r = owner().displayBox();
        return new Rectangle(r.x, r.y, 0, r.height);
    }
}