/*
 * Created on Apr 26, 2003
 */
package de.renew.diagram.peer;

import CH.ifa.draw.figures.AttributeFigure;
import CH.ifa.draw.figures.TextFigure;
import CH.ifa.draw.framework.ChildFigure;
import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.DrawingView;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureEnumeration;
import CH.ifa.draw.standard.StandardDrawingView;

import de.renew.diagram.DiagramTextFigure;
import de.renew.diagram.IDiagramElement;
import de.renew.diagram.MessageConnection;
import de.renew.diagram.PaletteCreatorPlugin;
import de.renew.diagram.RightConnector;
import de.renew.diagram.TailFigure;
import de.renew.diagram.TaskFigure;
import de.renew.diagram.VJoinFigure;
import de.renew.diagram.VSplitFigure;
import de.renew.gui.ArcConnection;
import de.renew.gui.PlaceFigure;
import de.renew.gui.TransitionFigure;
import de.renew.netcomponents.NetComponentFigure;

import java.awt.Point;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;


/**
 * <p>
 *     DrawingPeer is a default implementation of the notion of the DrawingPeer relation as defined by {@link IDrawingPeer}
 *     for transforming AIP drawing objects into Petri Net templates and connecting them.
 * </p>
 * <p>
 *     In Addition to the interface implementation, this class provides default X- and Y-offset values for spacing
 *     components in the Petri Net drawing.
 * </p>
 *  @author Lawrence Cabac
 */
public class DrawingPeer implements Serializable, IDrawingPeer {
    /**
     * The all-instances logger for this Class
     */
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(DrawingPeer.class);
    /**
     * The Vector of Petri Net Figures which are part of this DrawingPeer's template
     */
    protected Vector<Figure> figuresV;
    /**
     * The location of the center of the box surrounding all members of DrawingPeer.
     */
    protected Point location;
    /**
     * The (x,y) canvas origin locations connectors from this Peer's member figures point to
     */
    protected Vector<Point> nextLocations;
    /**
     * Vector of all Petri Net figures which require incoming arc connections from other DrawingPeers for
     * connecting this DrawingPeer's subnet to other DrawingPeers subnets.
     */
    protected Vector<Figure> startFigures;
    /**
     * Vector of all Petri Net figures which require outgoing arc connections to other DrawingPeers for
     * connecting this DrawingPeer's subnet to other DrawingPeers subnets.
     */
    protected Vector<Figure> endFigures;
    /**
     * Default X offset for layouting of figures
     */
    static final int XOFFSET = 300;
    /**
     * Default Y offset for layouting of figures
     */
    static final int YOFFSET = 200;
    /**
     * The TailFigure which specifically owns this peer
     */
    private TailFigure owner;
    /**
     * The String name of this DrawingPeer, identifying its template name.
     */
    private String name;

    /**
     * Default constructor, taking the owner, name and location.
     *
     * @param owner The TailFigure owning and originating this DrawingPeer
     * @param location The location the center of the peer is placed in the drawing
     * @param fileName The name of the peer as String, identifying its template.
     */
    public DrawingPeer(TailFigure owner, Point location, String fileName) {
        this.name = fileName;
        this.owner = owner;
        this.location = location;
    }

    /**
     * Empty constructor for use in {@link de.renew.dcdiagram.DCDrawingPeer}
     */
    public DrawingPeer() {
    }

    /**
     * Creates a new peer with the given name at the given location.
     * The owner is the aip-figure from which the peer is created.
     *
     * @param location The location the center of the peer is placed in the drawing
     * @param fileName The name of the peer as String, identifying its template
     * @param owner The TailFigure owning and originating this DrawingPeer
     * @param originator The originator of the peer as IDiagramElement
     */
    public DrawingPeer(Point location, String fileName, TailFigure owner,
                       IDiagramElement originator) {
        this.owner = owner;
        this.location = location;
        this.name = fileName;
        nextLocations = new Vector<Point>();

        // this variable should be passed as parameter (default = false) 
        boolean isSetIDs = true;

        NCLoader io = PaletteCreatorPlugin.getNCLoader();
        figuresV = new Vector<Figure>();
        FigureEnumeration enumeration = io.getFigures(fileName);
        while (enumeration != null && enumeration.hasMoreElements()) {
            figuresV.add(enumeration.nextElement());
        }

        if (isSetIDs) {
            String theTextID = "";

            // get the Task-ID from the Owner 
            FigureEnumeration children = owner.children();

            // TextFigure for the ID
            DiagramTextFigure theIDTextFigure = new DiagramTextFigure();
            int counter = 0;
            while (children.hasMoreElements()) {
                Figure figure = children.nextElement();
                if (figure instanceof TextFigure) {
                    TextFigure theTextFigure = (TextFigure) figure;
                    String text = theTextFigure.getText();
                    if (text.startsWith("##ID")) {
                        counter++;
                        theTextID += text;
                    }
                }
            }

            logger.debug("Number of IDs in DrawingPeer: " + counter);

            // get the Message-ID
            if (originator instanceof MessageConnection) {
                MessageConnection theMessage = (MessageConnection) originator;
                FigureEnumeration messageChildren = theMessage.children();
                while (messageChildren.hasMoreElements()) {
                    Figure figure = messageChildren.nextElement();
                    if (figure instanceof TextFigure) {
                        TextFigure theMessageTextFigure = (TextFigure) figure;
                        String text = theMessageTextFigure.getText();
                        if (text.startsWith("##ID")) {
                            theTextID += text;
                        }
                    }
                }
            }

            // set the TextID in the TextFigure
            theIDTextFigure.setText(theTextID);


            // attach the Task-ID to the NC
            Iterator<Figure> iterator = figuresV.iterator();
            while (iterator.hasNext()) {
                Figure figure = iterator.next();
                if (figure instanceof NetComponentFigure) {
                    NetComponentFigure theNCFigure = (NetComponentFigure) figure;
                    theNCFigure.attach(theIDTextFigure);
                    break;
                }
            }
        }

        startFigures = findStartConnectionTransitions(figuresV);
        endFigures = findEndConnectionPlaces(figuresV);
        int plusx = 0;
        int plusy = 0;


        // if (fileName.equals("start.rnw")) {
        // plusx = 150;
        // plusy = 300;
        // }
        if (this.getOwner() instanceof VSplitFigure) {
            logger.debug("===============================================");
            logger.debug("VSplitFigure");
            logger.debug("===============================================");

            plusy = YOFFSET;
            nextLocations.add(new Point(location.x + XOFFSET + plusx,
                            location.y + plusy));
            nextLocations.add(new Point(location.x + XOFFSET + plusx,
                            location.y - plusy));

            if ("cond.rnw".equals(name)) {
                logger.debug("DrawingPeer: encountered cond component.");
                TextFigure textFigure = getSelectedInscriptionFigure(figuresV,
                                "$booleanCondition");
                if ((textFigure != null) && (this.getOwner().children()
                                .hasMoreElements())) {
                    DiagramTextFigure dtf = (DiagramTextFigure) this.getOwner()
                                    .children().nextFigure();
                    String actionText = dtf.getRealText();

                    if (actionText != null && !actionText.startsWith("##")) {
                        logger.debug("  DrawingPeer: replaced text: "
                                        + actionText);
                        if (actionText.matches("[.*]")) {
                            actionText = actionText.substring(1,
                                            actionText.length() - 2);
                        }
                        if (!actionText.startsWith("cond =")) {
                            actionText = "cond = " + actionText;
                        }
                        textFigure.setText(actionText);
                    }
                }
            }
        } else if (this.getOwner() instanceof VJoinFigure) {
            logger.debug("===============================================");
            logger.debug("VJoinFigure");
            logger.debug("===============================================");


            nextLocations.add(new Point(location.x + XOFFSET + plusx,
                            location.y + plusy));


        } else if (this.getOwner() instanceof TaskFigure) {
            if (logger.isDebugEnabled()) {
                logger.debug("\n===============================================");
                logger.debug("TaskFigure");
                logger.debug("===============================================\n");
            }
            if (location == null) {
                location = new Point(0, 0);
            }
            int extra = 0;
            if (name.endsWith("exchange.rnw")) {
                extra = 100;
            }
            nextLocations.add(new Point(location.x + XOFFSET + plusx + extra,
                            location.y + plusy));
            if ("out.rnw".equals(name)) {
                logger.debug("DrawingPeer: encountered out component.");
                TextFigure textFigure = getSelectedInscriptionFigure(figuresV,
                                "$generateMessage");
                if (textFigure != null) {
                    logger.debug("  DrawingPeer: found textFigure with text: $generateMessage");
                    if (originator instanceof MessageConnection) {
                        MessageConnection message = (MessageConnection) originator;
                        if (message.getMessageText() != null && !message
                                        .getMessageText().startsWith("##")) {
                            if (message.getMessageText().startsWith("action")) {
                                // Text in diagram starts with keyword action, we presume
                                // that this is Java code:
                                textFigure.setText(message.getMessageText());
                                logger.debug("  DrawingPeer: replaced text: "
                                                + message.getMessageText());

                            } else {
                                // Otherwise we presume it is a message content (standard settings request)
                                textFigure.setText(
                                                "action p2 = Sl0Creator.createActionRequest(\n   aid,\n   "
                                                                + message.getMessageText()
                                                                + ")");
                                logger.debug("  DrawingPeer: replaced text: "
                                                + "action p2 = Sl0Creator.createActionRequest(\n   aid,\n   "
                                                + message.getMessageText()
                                                + ")");


                            }
                        } else {
                            textFigure.setText(
                                            "action p2=Sl0Creator.createActionRequest(aid,\"content\")");
                        }
                    } else {
                        logger.warn("DrawingPeer: Could not find text figure for outgoing message inscription.");
                    }
                }
            } else if ("sequence-template.rnw".equals(name)) {
                logger.debug("DrawingPeer: encountered sequence component.");
                TextFigure textFigure = getSelectedInscriptionFigure(figuresV,
                                "$sequenceAction");
                if (textFigure != null) {
                    DiagramTextFigure dtf = (DiagramTextFigure) originator;
                    String actionText = dtf.getRealText();
                    Object attr = dtf.getAttribute(DiagramTextFigure.IS_HIDDEN);
                    if (attr != null && (Boolean) attr) {
                        actionText = (String) dtf.getAttribute(
                                        DiagramTextFigure.HIDDEN_TEXT);
                    }
                    logger.debug("  DrawingPeer: replaced text: " + actionText);
                    textFigure.setText(":access(kb);\n" + actionText);
                }
            } else if ("stop.rnw".equals(name)) {
                logger.debug("DrawingPeer: encountered stop component.");
                TextFigure textFigure = getSelectedInscriptionFigure(figuresV,
                                "$sequenceAction");
                if (textFigure != null) {
                    DiagramTextFigure dtf = (DiagramTextFigure) originator;
                    String actionText = dtf.getRealText();
                    Object attr = dtf.getAttribute(DiagramTextFigure.IS_HIDDEN);
                    if (attr != null && (Boolean) attr) {
                        actionText = (String) dtf.getAttribute(
                                        DiagramTextFigure.HIDDEN_TEXT);
                    }
                    logger.debug("  DrawingPeer: replaced text: " + actionText);
                    if (actionText.equals("stop")
                                    || actionText.equals(":stop()")) {
                        textFigure.setText(":stop()");
                    }
                }
            } else if ("manual.rnw".equals(name)) {
                logger.debug("DrawingPeer: encountered stop component.");
                TextFigure textFigure = getSelectedInscriptionFigure(figuresV,
                                "$sequenceAction");
                if (textFigure != null) {
                    DiagramTextFigure dtf = (DiagramTextFigure) originator;
                    String actionText = dtf.getRealText();
                    Object attr = dtf.getAttribute(DiagramTextFigure.IS_HIDDEN);
                    if (attr != null && (Boolean) attr) {
                        actionText = (String) dtf.getAttribute(
                                        DiagramTextFigure.HIDDEN_TEXT);
                    }
                    logger.debug("  DrawingPeer: replaced text: " + actionText);
                    if (actionText.equals("manual")
                                    || actionText.equals("manual;")) {
                        textFigure.setText("manual");
                    }
                }
            } else if (name.endsWith("exchange.rnw")) {
                logger.debug("DrawingPeer: encountered exchange component.");
                TextFigure textFigure = getSelectedInscriptionFigure(figuresV,
                                "$descr");
                if (textFigure != null) {
                    DiagramTextFigure dtf = (DiagramTextFigure) originator;
                    String actionText = dtf.getRealText();
                    Object attr = dtf.getAttribute(DiagramTextFigure.IS_HIDDEN);
                    if (attr != null && (Boolean) attr) {
                        actionText = (String) dtf.getAttribute(
                                        DiagramTextFigure.HIDDEN_TEXT);
                    }
                    if (actionText.startsWith("simple ")) {
                        actionText = actionText.substring(7);
                    } else if (actionText.startsWith("simple")) {
                        actionText = actionText.substring(6);
                    }
                    logger.debug("  DrawingPeer: replaced text: " + actionText);
                    textFigure.setText(actionText);
                }
            }
        } else {
            nextLocations.add(new Point(location.x + XOFFSET + plusx,
                            location.y + plusy));
        }
        logger.debug("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
        logger.debug("Name " + fileName);
        logger.debug("location " + location);
        logger.debug("nextLoc " + nextLocations.firstElement());
        logger.debug(" ");
        logger.debug(" ");
        logger.debug(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");


    }

    private TextFigure getSelectedInscriptionFigure(Vector<Figure> figuresV,
                                                    String string) {
        return getSelectedInscription(figuresV, string);
    }


    @Override
    public Vector<Figure> getFigures() {
        return figuresV;
    }

    @Override
    public Point getLocation() {
        return location;
    }

    @Override
    public void setLocation(Point point) {
        location = point;
    }

    @Override
    public Vector<Point> getNextLocations() {
        return nextLocations;
    }

    @Override
    public Vector<Figure> getStartFigures() {
        return startFigures;
    }

    @Override
    public Vector<Figure> getEndFigures() {
        return endFigures;
    }

    @Override
    public void drawFigures(DrawingView view) {
        view.addAll(figuresV);
        view.addToSelectionAll(figuresV);
        StandardDrawingView.moveFigures(figuresV, getLocation().x,
                        getLocation().y);
        view.clearSelection();
        view.checkDamage();
    }

    @Override
    public void connectFigures(DrawingView view) {
        if (!this.toString().equals("start.rnw")) {
            connectToPreceding(view);
            view.checkDamage();
        }
    }

    /**
     * Connects the current peer to the preceding. In case the peer is a join, it connects to both preceding peers.
     * TODO add Unit-Tests
     *
     * @param view The current DrawingView, i.e. the view associated to the DiagramDrawing.
     */
    private void connectToPreceding(DrawingView view) {
        AttributeFigure start = null;
        Vector<Connector> v = this.getOwner().getDParentConnectors(); // Connectors of the parent

        // diagram figures
        int index = 0;
        if (v.size() > 0) {
            Connector con = v.firstElement();
            if (con instanceof RightConnector) {
                index++;
            }


            // get the correct place marked with a ">" of the last peer of the
            // parent
            Vector<Figure> endFigures2 = this.getOwner().getPrecedingPeer(this)
                            .getEndFigures();
            try {
                start = (AttributeFigure) endFigures2.elementAt(index);
            } catch (ArrayIndexOutOfBoundsException e) {
                // logger.error(e.getMessage(), e);
                if (endFigures2.size() > 0) {
                    start = (AttributeFigure) endFigures2.elementAt(0);
                }
            }
            if (start != null) {
                AttributeFigure end;
                ArcConnection connection;

                if (this.getOwner() instanceof VJoinFigure) {
                    AttributeFigure start2;
                    AttributeFigure end2;

                    Vector<AttributeFigure> connections = getConnectionElementsJoin();
                    start = connections.elementAt(0);
                    start2 = connections.elementAt(1);
                    end = connections.elementAt(2);
                    end2 = connections.elementAt(3);

                    //add connections to view
                    connection = new ArcConnection(1);
                    connection.startPoint(start2.center());
                    connection.endPoint(end.center());
                    connection.connectEnd(end.connectorAt(end.center()));
                    connection.connectStart(
                                    start2.connectorAt(start2.center()));
                    logger.debug("connection " + connection);
                    connection.updateConnection();
                    view.add(connection);

                    ArcConnection connection2 = new ArcConnection(1);
                    connection2.startPoint(start2.center());
                    connection2.endPoint(end2.center());
                    connection2.connectEnd(end2.connectorAt(end2.center()));
                    connection2.connectStart(start.connectorAt(start.center()));
                    logger.debug("connection " + connection2);
                    connection2.updateConnection();
                    view.add(connection2);

                } else {
                    // get the correct transitions of this drawing peer
                    end = (AttributeFigure) getStartFigures().firstElement();
                    connection = new ArcConnection(1);


                    connection.startPoint(start.center());
                    connection.endPoint(end.center());
                    connection.connectEnd(end.connectorAt(end.center()));
                    connection.connectStart(start.connectorAt(start.center()));
                    logger.debug("connection " + connection);
                    connection.updateConnection();
                    view.add(connection);
                }
            }
        }
    }

    /**
     * For the ending Petri Net figures of this Peer's subnet, update the stored canvas coordinate locations
     * based on the figure center and the default offset.
     */
    protected void updateNextLocations() {
        Iterator<Figure> it = endFigures.iterator();
        while (it.hasNext()) {
            AttributeFigure fig = (AttributeFigure) it.next();
            Point nextLocation = new Point(fig.center().x + XOFFSET,
                            fig.center().y);
            nextLocations.add(nextLocation);
        }
    }

    /**
     * For the submitted set of Petri Net figures, identify those figures which are {@link TransitionFigure} and whose name
     * is configured as the correct identifier and thus considered starting connection transitions.
     * @param figures the Figures for which to check for StartingFigures
     * @return the subset of Figures which actually are starting transition figures.
     */
    protected Vector<Figure> findStartConnectionTransitions(Vector<Figure> figures) {
        Iterator<Figure> it = figures.iterator();
        Vector<Figure> result = new Vector<Figure>();
        while (it.hasNext()) {
            Figure fig = it.next();
            if (fig instanceof TransitionFigure) {
                if (figureIsNamed(fig, ">")) {
                    result.add(fig);
                    logger.debug("################## Start connectable figures "
                                    + fig);
                }
            }
        }

        return result;
    }

    /**
     * For the submitted set of Figures, identify those figures which are {@link PlaceFigure} and whose name is configured
     * as the correct identifier and thus considered ending connection places.
     * @param figures the Figures for which to check for ending connection places
     * @return the subset of Figures which actually are ending connection places
     */
    protected Vector<Figure> findEndConnectionPlaces(Vector<Figure> figures) {
        Iterator<Figure> it = figures.iterator();
        Vector<Figure> result = new Vector<Figure>();
        while (it.hasNext()) {
            Figure fig = it.next();
            if (fig instanceof PlaceFigure) {
                if (figureIsNamed(fig, ">")) {
                    result.add(fig);
                    logger.debug("################## Start connectable figures "
                                    + fig);
                }
            }
        }

        return result;
    }

    //TODO add Unit-Tests
    /**
     * Retrieve the AttributeFigures which should be connected for a Joining relation. These are the starting transitions on the join
     * and the ending places of both preceding AIP figure's respective peers.
     * @return the AttributeFigures to connect according to specification
     */
    private Vector<AttributeFigure> getConnectionElementsJoin() {
        Vector<AttributeFigure> connections = new Vector<>();

        AttributeFigure start;
        AttributeFigure start2;
        AttributeFigure end;
        AttributeFigure end2;

        AttributeFigure startFigure1 = (AttributeFigure) getStartFigures().elementAt(0);
        AttributeFigure startFigure2 = (AttributeFigure) getStartFigures().elementAt(1);
        //In a VJoinFigure the preceding and the second preceding elements are relevant, because both need to be connected to the join
        AttributeFigure endFigure1 = (AttributeFigure) this.getOwner().getPrecedingPeer(this).getEndFigures().elementAt(0);
        AttributeFigure endFigure2 = (AttributeFigure) this.getOwner().getSecondPrecedingPeer(this).getEndFigures().elementAt(0);

        //the lower component is the LeftConnector of the Join
        if (startFigure1.center().y > startFigure2.center().y) {
            end = startFigure2;
            end2 = startFigure1;
        }
        else {
            end = startFigure1;
            end2 = startFigure2;
        }

        //the lower component should connect to the LeftConnector of the Join
        if (endFigure1.center().y > endFigure2.center().y) {
            start = endFigure1;
            start2 = endFigure2;
        }
        else {
            start = endFigure2;
            start2 = endFigure1;
        }

        connections.add(start);
        connections.add(start2);
        connections.add(end);
        connections.add(end2);

        return connections;
    }

    /**
     * For a set of Figures, retrieve those Figures which are transition or place figures and have the inscription
     * specified by the string parameter
     * @param figures the List of figures for which to check
     * @param inscription the inscription valid figures should have
     * @return those figures in the submitted vector which have the matching string as their inscription
     */
    protected TextFigure getSelectedInscription(Vector<Figure> figures,
                                                String inscription) {
        Iterator<Figure> it = figures.iterator();
        TextFigure result = null;
        while (it.hasNext()) {
            Figure fig = it.next();
            if (fig instanceof TransitionFigure || fig instanceof PlaceFigure) {
                if (figureIsNamed(fig, inscription)) {
                    result = getNamedChild(fig, inscription);
                    logger.debug("################## Start connectable figures "
                                    + fig);
                }
            }
        }

        return result;
    }

    static private TextFigure getNamedChild(Figure fig, String str) {
        if (!(fig instanceof AttributeFigure)) {
            return null;
        }
        AttributeFigure attfig = (AttributeFigure) fig;
        Enumeration<Figure> enumeration = attfig.children();
        TextFigure result = null;
        while (enumeration.hasMoreElements()) {
            ChildFigure child = (ChildFigure) enumeration.nextElement();
            if (child instanceof TextFigure) {
                TextFigure textFigure = (TextFigure) child;
                if (textFigure.getText().equals(str)) {
                    result = textFigure;
                }
            }
        }
        return result;
    }

    static private boolean figureIsNamed(Figure fig, String str) {
        if (!(fig instanceof AttributeFigure)) {
            return false;
        }
        AttributeFigure attfig = (AttributeFigure) fig;
        Enumeration<Figure> enumeration = attfig.children();
        boolean result = false;
        while (enumeration.hasMoreElements()) {
            ChildFigure child = (ChildFigure) enumeration.nextElement();
            if (child instanceof TextFigure) {
                TextFigure textFigure = (TextFigure) child;
                if (textFigure.getText().equals(str)) {
                    result = true;
                }
            }
        }
        return result;
    }

    /**
     * Retrieve the next canvas x,y location for the given Connector. This is either the
     * first of the next possible locations or, if this DrawingPeer's TailFigure owner has a
     * parent connector and that connector is for the right side of a {@link VSplitFigure}, the
     * second configured Location.
     * @param con the Connector for which to check
     * @return the canvas location for the given Connector to point at
     */
    @Override
    public Point getNextLocation(Connector con) {
        // return (Point)getNextLocations().lastElement();
        Point result = null;
        Vector<Connector> v = this.getOwner().getDParentConnectors();
        int index = 0;
        if (!v.isEmpty()) {
            logger.debug("con " + con);


            // check whether the owner is connected to the right side of a
            // VSplitFigure
            if (con instanceof RightConnector) {
                index++;
            }
        }
        result = getNextLocations().elementAt(index);
        logger.debug("index " + index);
        logger.debug("result " + result);
        return result;


        // AttributeFigure interfaceState = null;
        // int x = XOFFSET;
        // int y = 0;
        // Vector v = owner.getDParentConnectors();
        // int index = 0;
        // if (v.size() > 0){
        // Connector con = (Connector) v.firstElement();
        // //check whether the owner is connected to the right side of a
        // VSplitFigure
        // y = - YOFFSET;
        // if (con instanceof RightConnector){
        // index++;
        // y = YOFFSET;
        // }
        // }
        // interfaceState = (AttributeFigure) getEndFigures().elementAt(index);
        // x += interfaceState.center().x;
        // y += interfaceState.center().y ;
        //        
        //        
        // return new Point(x,y);
    }


    /**
     * Return the String name of this DrawingPeer
     * @return the String name configured for this DrawingPeer at construction.
     */
    @Override
    public String toString() {
        return name;
    }

    @Override
    public TailFigure getOwner() {
        return owner;
    }
}