package de.renew.diagram.peer;

import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.DrawingView;
import CH.ifa.draw.framework.Figure;

import de.renew.diagram.TailFigure;

import java.awt.Point;
import java.util.Vector;

/**
 * <p>
 * DrawingPeers are the Petri Net-level representations of individual AIP figure objects such as {@link TailFigure} and
 * {@link de.renew.diagram.DiagramFigure} objects.
 * Using DrawingPeers, the AIP net parts are translated further into templates of Petri net components which are connected together using
 * the connection hierarchies of the represented AIP diagram objects.
 * </p>
 * <p>
 * Towards this end each DrawingPeer provides a vector of Petri Net figures for drawing that Petri net, is capable of locating itself on a
 * drawing canvas, provides identification of its starting and ending Petri net components for connection to the Petri net components of
 * other DrawingPeers, and provides hooks for executing the actual drawing of components on the canvas and their connection with each other.
 * </p>
 */
public interface IDrawingPeer {
    /**
     * Retrieve the Vector of all Petri Net figures which are part of its DrawingPeer's template.
     * @return the Vector of Petri Net figures which are specified for this DrawingPeer.
     */
    Vector<Figure> getFigures();

    /**
     * Retrieve the (x,y) canvas coordinates for the center of the Peer placement on the canvas
     * @return x,y canvas origin coordinates for the Peer center
     */
    Point getLocation();

    /**
     * Set the x,y canvas origin coordinates for the center of this Peer on the canvas
     * @param point the x,y canvas origin coordinates to set
     */
    void setLocation(Point point);

    /**
     * Retrieve a vector of all x,y canvas coordinates for the next Petri Net Figures to connect to on the canvas
     * @return a Vector of (x,y) coordinate points in canvas coordinates pointing to the next figures with which to create connections
     */
    Vector<Point> getNextLocations();

    // abstract public DiagramTextFigure inscription();

    /**
     * The Start Figures of a Peer are those Petri Net components which require incoming connections from other DrawingPeer templates
     * in order for template parts to be connected together according to the AIP drawing. This function returns those specific figures
     * which are part of this DrawingPeer.
     * @return the Figures which are starting figures for this DrawingPeer, receiving incoming arcs from other Peers.
     */
    Vector<Figure> getStartFigures();

    /**
     * The End figures of a peer connection are those Petri Net components which lead int outgoing connections to other
     * DrawingPeer templates in order for template parts to be connected together according to the AIP drawing. This function
     * returns those specific figures which are part of this DrawingPeer.
     * @return the Figures which are ending figures for this DrawingPeer, connecting outgoing arcs to other Peers.
     */
    Vector<Figure> getEndFigures();

    /**
     * All figures in this peer will be drawn then selected and moved to their
     * position. Then a connection between the preceding and these figures will
     * be created.
     *
     * @param view The view to draw the figures in.
     */
    void drawFigures(DrawingView view);

    /**
     * All figures in this peer will be connected together in their configured relationships.
     * @param view the view to connect the figures in
     */
    void connectFigures(DrawingView view);

    /**
     * For a given Connector object, retrieve the next possible connection location for this Connector
     * in canvas origin coordinates.
     * @param con the Connector for which to check
     * @return the (x,y) Canvas coordinates of the next location for the submitted Connector
     */
    Point getNextLocation(Connector con);

    @Override
    String toString();

    /**
     * Retrieve the TailFigure which owns this specific Peer relationship.
     * @return the owning TailFigure object.
     */
    TailFigure getOwner();
}