package diagram;

import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.Figure;
import de.renew.diagram.HBottomConnector;
import de.renew.diagram.HSplitCenterConnector;
import de.renew.diagram.HSplitFigure;
import de.renew.diagram.HTopConnector;
import de.renew.diagram.HorizontalConnector;
import de.renew.diagram.MessageConnection;
import de.renew.diagram.MessageConnectionHandle;
import de.renew.diagram.TaskFigure;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;

import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;

/**
 * This class is a test class for {@link MessageConnectionHandle}.
 * It tests if the right messages are called and the right connector is found.
 *
 * @author Lukas Seifert
 */
public class MessageConnectionHandleTest {

    /**
     * This inner class provides locally scoped access to protected methods of {@link MessageConnectionHandle}.
     */
    static class TestMessageConnectionHandle extends MessageConnectionHandle {

        /**
         * Constructs a handle for the given owner. Uses a default center location in the owner figure
         * and the {@link MessageConnection} message type.
         *
         * @param owner the Figure which owns this handle and from which new connections will be created.
         */
        public TestMessageConnectionHandle(Figure owner) {
            super(owner);
        }

        @Override
        protected Connector findConnector(int x, int y, Figure f) {
            return super.findConnector(x, y, f);
        }

        @Override
        protected Connector startConnector() {
            return super.startConnector();
        }
    }

    @Mock
    private TaskFigure taskOwner;
    private TestMessageConnectionHandle testMessageConnectionHandle;

    @BeforeEach
    public void setUp() {
        taskOwner = mock(TaskFigure.class);
        testMessageConnectionHandle = spy(new TestMessageConnectionHandle(taskOwner));
    }

    /**
     * This method tests the method findConnector from {@link MessageConnectionHandle}.
     * It ensures that if the given figure is a TaskFigure the method {@link TaskFigure#messageConnectorAt()}
     * is called and a HorizontalConnector is returned.
     * If the given figure is a HSplitFigure the method {@link HSplitFigure#connectorAt(int, int)}
     * shall be called and a HBottomConnector, HTopConnector or HSplitCenterConnector returned, belonging on the coordinates.
     * In the upper part on the default box created at (0,0) this is a HTopConnector,
     * in the lower part a HBottomConnector and in the middle a HSplitCenterConnector.
     */
    @Test
    public void findConnectorTest() {
        TaskFigure testTaskFigure = spy(TaskFigure.class);
        HSplitFigure testHSplitFigure = spy(HSplitFigure.class);

        Connector connector = testMessageConnectionHandle.findConnector(5, 5, testHSplitFigure);
        assertTrue(connector instanceof HBottomConnector);
        verify(testHSplitFigure).connectorAt(5, 5);

        connector = testMessageConnectionHandle.findConnector(-5, -5, testHSplitFigure);
        assertTrue(connector instanceof HTopConnector);
        verify(testHSplitFigure).connectorAt(-5, -5);

        connector = testMessageConnectionHandle.findConnector(0, 0, testHSplitFigure);
        assertTrue(connector instanceof HSplitCenterConnector);
        verify(testHSplitFigure).connectorAt(0, 0);

        connector = testMessageConnectionHandle.findConnector(5, 5, testTaskFigure);
        assertTrue(connector instanceof HorizontalConnector);
        verify(testTaskFigure).messageConnectorAt();
    }

    /**
     * This method tests the method startConnector from {@link MessageConnectionHandle}.
     * It ensures the method {@link TaskFigure#messageConnectorAt()} is called if the owner is a TaskFigure.
     */
    @Test
    public void startConnectorTest() {
        testMessageConnectionHandle.startConnector();
        verify(taskOwner).messageConnectorAt();
    }

}