package de.renew.plugin.load;

import java.lang.module.ModuleReference;
import java.util.Collection;
import java.util.List;

/**
 * This exception is thrown by the module manager when it fails to load a
 * module due to conflicts in the loading order.
 */
public class ModuleLayerCreationException extends RuntimeException {
    /**
     * A constructor for the ModuleLayerCreationException.
     * Creates an exception message when more than one module is included in a layer
     * @param moduleToLoad A string with the loaded modules descriptor
     * @param loadedModules A collection of modules that were added to the layer that caused the exception.
     */
    public ModuleLayerCreationException(String moduleToLoad, Collection<Module> loadedModules) {
        super("While loading the module " + moduleToLoad + ", " +
                "the Java Module System attempted to include one or more different modules in the same layer." +
                "This is usually the result of an incorrect loading order of plugins. " +
                "The following modules were included in this layer: " +
                loadedModules.stream().map(Module::getName).reduce("", String::concat) + " " +
                "Please make sure that the additional modules are loaded before " + moduleToLoad + ".");
    }

    /**
     * A constructor for the ModuleLayerCreationException.
     * Creates an exception if parent layers are not located during Module load.
     * @param moduleToLoad The Module that was attempted to load
     * @param requiredLayers A String collection containing names of required parent layers to load the Module
     * @param foundLayers A Module Collection of parent layers that were located during Module load
     */
    public ModuleLayerCreationException(ModuleReference moduleToLoad,
                                        List<String> requiredLayers,
                                        List<ModuleLayer> foundLayers) {
        super("Couldn't find all of the required parent layers for module " + moduleToLoad.descriptor().name() + ". " +
                "Required parent layers: " + requiredLayers.toString() + ". " +
                "Found layers: " + foundLayers.toString() + ". " +
                "This is usually the result of an incorrect loading order of plugins. " +
                "Please make sure that all required modules are loaded before " + moduleToLoad.descriptor().name()
                + ".");
    }
}
