package de.renew.ptchannel.single;

import de.renew.formalism.java.InscriptionParser;
import de.renew.formalism.java.JavaNetParser;
import de.renew.formalism.java.ParseException;
import de.renew.formalism.java.ParsedDeclarationNode;
import de.renew.formalism.pt.SinglePTNetCompiler;
import de.renew.net.TransitionInscription;
import de.renew.shadow.*;

import java.io.StringReader;
import java.util.Collection;

/**
 * This is the single for P/T nets with synchronous channels.
 * For a definition, see thesis of Voß or Kühnemund (2021).
 * It behaves like a P/T net single and only overrides methods
 * that are necessary for channels.
 * Channels are compiled using a specialized {@link PTCTransitionCompiler}.
 * Importantly, transitions may only carry one inscription.
 *
 * Integer parameters for channels are valid. They are compiled by
 * converting them into lists with n black tokens and converting
 * all corresponding inscripted arcs into flexible arcs which
 * unpack those lists.
 *
 * @author Lukas Voß
 */
public class SinglePTNetWithChannelCompiler extends SinglePTNetCompiler {

    protected final PTCTransitionCompiler transitionCompiler;
    protected ParsedDeclarationNode declaration;

    /**
     * Creates a new <code>SinglePTNetWithChannelCompiler</code> instance.
     * The only additional step to the super constructor is setting
     * a {@link PTCTransitionCompiler} for compilation of transition
     * inscriptions.
     */
    public SinglePTNetWithChannelCompiler() {
        super();
        this.transitionCompiler = new PTCTransitionCompiler(false, false, false);
    }

    /**
     * This is the syntax check for transition inscriptions.
     * After a transition inscription is edited, this method gets
     * called. It throws an exception whenever the inscription
     * includes anything other than an up-/downlink.
     *
     * @param inscription the edited entire inscription
     * @param special     whether it is a special inscription
     * @param shadowNet   the corresponding shadow net
     * @return type of the first inscription
     * @throws SyntaxException whenever an inscription is no up-/downlink
     */
    @Override
    public String checkTransitionInscription(String inscription,
                                             boolean special,
                                             ShadowNet shadowNet)
            throws SyntaxException {
        String trans = transitionCompiler.checkTransitionInscription(inscription, special, shadowNet);
        declaration = transitionCompiler.makeDeclarationNode(shadowNet);
        return trans;
    }

    @Override
    public String checkDeclarationNode(String inscr, boolean special,
                                       ShadowNet shadowNet)
            throws SyntaxException {
        parseDeclarationNode(inscr);
        return "declaration";
    }

    ParsedDeclarationNode parseDeclarationNode(String inscr)
            throws SyntaxException {
        if (inscr != null) {
            InscriptionParser parser = makeParser(inscr);
            try {
                return parser.DeclarationNode();
            } catch (ParseException e) {
                throw PTCTransitionCompiler.makeSyntaxException(e);
            }
        } else {
            return makeEmptyDeclarationNode(null);
        }
    }

    protected ParsedDeclarationNode makeEmptyDeclarationNode(ShadowNet net) {
        return new ParsedDeclarationNode();
    }

    @Override
    public void parseDeclarations(ShadowNet shadowNet) throws SyntaxException {
        makeDeclarationNode(shadowNet);
    }

    public ParsedDeclarationNode makeDeclarationNode(ShadowNet shadowNet)
            throws SyntaxException {
        // Compile the declaration node.
        ShadowDeclarationNode node = findDeclarationNode(shadowNet);
        if (node == null) {
            declaration = makeEmptyDeclarationNode(shadowNet);
        } else {
            declaration = compile(node);
        }
        return declaration;
    }

    protected ParsedDeclarationNode compile(ShadowDeclarationNode declaration)
            throws SyntaxException {
        try {
            return parseDeclarationNode(declaration.inscr);
        } catch (SyntaxException e) {
            throw e.addObject(declaration);
        }
    }

    /**
     * When starting a simulation, this method gets called to compile
     * transition inscriptions.
     * The compilation process takes place in the {@link PTCTransitionCompiler}.
     *
     * @param inscription inscription to compile
     * @return All inscriptions of a transition
     * @throws SyntaxException whenever a inscription is no up-/downlink
     */
    @Override
    protected Collection<TransitionInscription> compileTransitionInscription(ShadowInscription inscription)
            throws SyntaxException {
        return transitionCompiler.makeInscriptions(inscription, lookup, true, declaration, loopbackNetLoader);
    }

    protected InscriptionParser makeParser(String inscr) {
        JavaNetParser parser = new JavaNetParser(new StringReader(inscr));
        parser.setNetLoader(loopbackNetLoader);
        return parser;
    }

    @Override
    public String checkArcInscription(String inscr, boolean special,
                                      ShadowNet shadowNet)
            throws SyntaxException {
        return transitionCompiler.checkArcInscription(inscr, special, shadowNet);
    }

    /**
     * Compile an arc like in the ordinary P/T formalism.
     * If that throws an exception, the arc is inscribed.
     * Thus, it needs to be compiled using the
     * transitionCompiler instead.
     *
     * @param shadowArc the arc to compile
     * @throws SyntaxException if the transition single cannot compile the arc
     */
    @Override
    protected void compileArc(ShadowArc shadowArc) throws SyntaxException {
        try {
            super.compileArc(shadowArc);
        } catch (SyntaxException e) {
            if (shadowArc.shadowArcType == ShadowArc.ordinary) {
                transitionCompiler.compileArc(shadowArc, declaration, lookup);
            }
            else {
                throw new SyntaxException("Only ordinary arcs are allowed.").addObject(shadowArc);
            }
        }
    }
}
